"use strict";
/*
 * Deepkit Framework
 * Copyright (C) 2021 Deepkit UG, Marc J. Schmidt
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the MIT License.
 *
 * You should have received a copy of the MIT License along with this program.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.PostgresPlatform = exports.PostgresPlaceholderStrategy = void 0;
const sql_1 = require("@deepkit/sql");
const postgres_serializer_1 = require("./postgres-serializer");
const type_1 = require("@deepkit/type");
const postgres_schema_parser_1 = require("./postgres-schema-parser");
const sql_filter_builder_1 = require("./sql-filter-builder");
const core_1 = require("@deepkit/core");
const sqlstring_1 = __importDefault(require("sqlstring"));
function escapeLiteral(value) {
    if (value === null || value === undefined)
        return 'null';
    if (value instanceof Date)
        return sqlstring_1.default.escape(value);
    if ('number' === typeof value || 'bigint' === typeof value)
        return String(value);
    if ('string' !== typeof value)
        return escapeLiteral(String(value));
    let hasBackslash = false;
    let escaped = '\'';
    for (let i = 0; i < value.length; i++) {
        const c = value[i];
        if (c === '\'') {
            escaped += c + c;
        }
        else if (c === '\\') {
            escaped += c + c;
            hasBackslash = true;
        }
        else {
            escaped += c;
        }
    }
    escaped += '\'';
    if (hasBackslash) {
        escaped = ' E' + escaped;
    }
    return escaped;
}
class PostgresPlaceholderStrategy extends sql_1.SqlPlaceholderStrategy {
    getPlaceholder() {
        return '$' + (++this.offset);
    }
}
exports.PostgresPlaceholderStrategy = PostgresPlaceholderStrategy;
class PostgresPlatform extends sql_1.DefaultPlatform {
    constructor() {
        super();
        this.defaultSqlType = 'text';
        this.annotationId = 'postgres';
        this.serializer = postgres_serializer_1.postgresSerializer;
        this.schemaParserType = postgres_schema_parser_1.PostgresSchemaParser;
        this.placeholderStrategy = PostgresPlaceholderStrategy;
        this.addType(6 /* number */, 'double precision');
        this.addType(7 /* boolean */, 'boolean');
        this.addType(20 /* class */, 'jsonb');
        this.addType(25 /* array */, 'jsonb');
        this.addType(23 /* union */, 'jsonb');
        this.addType(v => v.kind === 22 /* enum */ && v.indexType.kind === 6 /* number */, 'integer');
        this.addType(v => v.kind === 22 /* enum */ && v.indexType.kind === 5 /* string */, 'text');
        this.addType(v => v.kind === 22 /* enum */ && v.indexType.kind === 23 /* union */, 'jsonb');
        this.addType(9 /* bigint */, 'bigint');
        this.addType(type => type.kind === 6 /* number */ && type.brand === type_1.TypeNumberBrand.integer, 'integer');
        this.addType(type => type.kind === 6 /* number */ && type.brand === type_1.TypeNumberBrand.int8, 'smallint');
        this.addType(type => type.kind === 6 /* number */ && type.brand === type_1.TypeNumberBrand.uint8, 'smallint');
        this.addType(type => type.kind === 6 /* number */ && type.brand === type_1.TypeNumberBrand.int16, 'smallint');
        this.addType(type => type.kind === 6 /* number */ && type.brand === type_1.TypeNumberBrand.uint16, 'smallint');
        this.addType(type => type.kind === 6 /* number */ && type.brand === type_1.TypeNumberBrand.int32, 'integer');
        this.addType(type => type.kind === 6 /* number */ && type.brand === type_1.TypeNumberBrand.uint32, 'integer');
        this.addType(type => type.kind === 6 /* number */ && type.brand === type_1.TypeNumberBrand.float32, 'real');
        this.addType(type => type.kind === 6 /* number */ && type.brand === type_1.TypeNumberBrand.float64, 'double precision');
        this.addType(type => type.kind === 6 /* number */ && type.brand === type_1.TypeNumberBrand.float, 'double precision');
        this.addType(type_1.isUUIDType, 'uuid');
        this.addBinaryType('bytea');
        this.addType(type => type.kind === 20 /* class */ && type.classType === Date, 'timestamp');
    }
    getAggregateSelect(tableName, property, func) {
        if (func === 'group_concat') {
            return `array_to_string(array_agg(${tableName}.${this.quoteIdentifier(property.name)}), ',')`;
        }
        return super.getAggregateSelect(tableName, property, func);
    }
    createSqlFilterBuilder(schema, tableName) {
        return new sql_filter_builder_1.PostgreSQLFilterBuilder(schema, tableName, this.serializer, new this.placeholderStrategy, this.quoteValue.bind(this), this.quoteIdentifier.bind(this));
    }
    quoteValue(value) {
        if (!(value instanceof Date) && ((0, core_1.isObject)(value) || (0, core_1.isArray)(value)))
            return escapeLiteral(JSON.stringify(value));
        if (value instanceof Date)
            return 'TIMESTAMP ' + sqlstring_1.default.escape(value);
        return escapeLiteral(value);
    }
    getColumnDDL(column) {
        const ddl = [];
        ddl.push(this.getIdentifier(column));
        if (column.isAutoIncrement) {
            ddl.push(`SERIAL`);
        }
        else {
            const foreignKey = column.table.getForeignKeyOfLocalColumn(column);
            if (foreignKey) {
                const [foreignPk] = foreignKey.foreignColumns;
                if (foreignPk.isAutoIncrement) {
                    //the foreignPK has `SERIAL` type, which is basically an INTEGER.
                    ddl.push('INTEGER');
                }
                else {
                    ddl.push((foreignPk.type || 'INTEGER') + column.getSizeDefinition());
                }
            }
            else {
                ddl.push((column.type || 'INTEGER') + column.getSizeDefinition());
            }
            ddl.push(column.isNotNull ? this.getNotNullString() : this.getNullString());
            ddl.push(this.getColumnDefaultValueDDL(column));
        }
        return ddl.filter(sql_1.isSet).join(' ');
    }
    getUniqueDDL(unique) {
        return `CONSTRAINT ${this.getIdentifier(unique)} UNIQUE (${this.getColumnListDDL(unique.columns)})`;
    }
    getDropIndexDDL(index) {
        return `DROP CONSTRAINT ${this.getIdentifier(index)}`;
    }
    supportsInlineForeignKey() {
        return false;
    }
    getAutoIncrement() {
        return '';
    }
    getDropTableDDL(table) {
        return `DROP TABLE IF EXISTS ${this.getIdentifier(table)} CASCADE`;
    }
    getUseSchemaDDL(table) {
        if (!table.schemaName)
            return '';
        return `SET search_path TO ${this.quoteIdentifier(table.schemaName)}`;
    }
    getResetSchemaDDL(table) {
        if (!table.schemaName)
            return '';
        return `SET search_path TO public`;
    }
}
exports.PostgresPlatform = PostgresPlatform;
//# sourceMappingURL=postgres-platform.js.map